#pragma once
#ifndef CERES_H
#define CERES_H

#include "/user/HS222/tw00275/ceres-bin/include/ceres/ceres.h"
#include "glog/logging.h"

using ceres::AutoDiffCostFunction;
using ceres::CostFunction;
using ceres::Problem;
using ceres::Solver;
using ceres::Solve;

void specularMinimisation(double& SpecularIntensity, double& SpecularPower, double& residualValue);

// A templated cost functor that implements the residual r = 10 - x.
// The method operator() is templated so that we can then use an automatic differentiation wrapper around it to generate its derivatives.
struct CostFunctor {
	
	// Pass variables into struct
	//CostFunctor(GLuint bufferID):bufferID(bufferID){};
	CostFunctor(double SpecularIntensity, double SpecularPower): SpecularIntensity(SpecularIntensity), SpecularPower(SpecularPower) {}
	
	template <typename T> bool operator()(const T* const SpecularIntensity, const T* const SpecularPower, T* residualValue) const {
		
		//residual[0] = x[0];
		residualValue[0] = SpecularIntensity[0];
		return true;
	}
	
	
	
	// Factory to hide the construction of the CostFunction object from the client code
	/*static ceres::CostFunction* Create(SpecularIntensity, SpecularPower, residualValue) {
		return (new ceres::AutoDiffCostFunction<CostFunctor, 1, 1, 1>(new CostFunctor(SpecularIntensity, SpecularPower, residualValue)));
	}*/
	
	
	// Access variables in struct
	private:
	//GLuint bufferID;
	double& SpecularIntensity;
	double& SpecularPower;
	//double& residualValue;
};

/*struct CostFunctor {
  template <typename T> bool operator()(const T* const SpecularIntensity, const T* const SpecularPower, T* residual) const {
    computeResiduals(residual, residuals, totalResidual, residualImage, calculateResidual, depthProjectionMatrix, depthViewMatrix, width, height, position, horizontalAngle, verticalAngle, FoV, lightInvDir, lightDirections, textureImages, lightNumber, numberOfLights, SpecularIntensityID, SpecularIntensity, SpecularPowerID, SpecularPower, programID, ModelMatrixID, ViewMatrixID, DepthBiasID, lightInvDirID, Texture, TextureID, depthTexture, ShadowMapID, vertexbuffer, uvbuffer, normalbuffer, elementbuffer, indices, MatrixID);
	
    residual[0] = residual;
    return true;
  }
};*/

void specularMinimisation(double& SpecularIntensity, double& SpecularPower, double& residualValue) {
	
  // The variable to solve for with its initial value. It will be mutated in place by the solver.
  const double initialSpecularIntensity = SpecularIntensity;
  const double initialSpecularPower     = SpecularPower;

  // Build the problem.
  ceres::Problem problem;

  // Set up the only cost function (also known as residual). This uses auto-differentiation to obtain the derivative (jacobian).
  // AutoDiffCostFunction<CostFunctor, (Dimensions of Residual), (Dimensions of Variables)>
  ceres::CostFunction* cost_function = new AutoDiffCostFunction<CostFunctor, 1, 1, 1>(new CostFunctor(SpecularIntensity, SpecularPower));
  //ceres::CostFunction* cost_function = CostFunctor::Create(SpecularIntensity, SpecularPower, residualValue);
  problem.AddResidualBlock(cost_function, NULL, &SpecularIntensity, &SpecularPower);

  // Run the solver!
  Solver::Options options;
  options.minimizer_progress_to_stdout = true;
  Solver::Summary summary;
  Solve(options, &problem, &summary);

  std::cout << summary.BriefReport() << "\n";
  std::cout << "Specular Intensity: " << initialSpecularIntensity << " -> " << SpecularIntensity << "\n";
  std::cout << "Specular Power: "     << initialSpecularPower     << " -> " << SpecularPower     << "\n";
}

/*void specularMinimisation(char** argv) {
	
  google::InitGoogleLogging(argv[0]);
  
  // The variable to solve for with its initial value. It will be
  // mutated in place by the solver.
  double x = 0.5;
  const double initial_x = x;

  // Build the problem.
  Problem problem;

  // Set up the only cost function (also known as residual). This uses
  // auto-differentiation to obtain the derivative (jacobian).
  CostFunction* cost_function = new AutoDiffCostFunction<CostFunctor, 1, 1>(new CostFunctor);
  problem.AddResidualBlock(cost_function, NULL, &x);

  // Run the solver!
  Solver::Options options;
  options.minimizer_progress_to_stdout = true;
  Solver::Summary summary;
  Solve(options, &problem, &summary);

  std::cout << summary.BriefReport() << "\n";
  std::cout << "x : " << initial_x << " -> " << x << "\n";
}*/

#endif
