#pragma once
#ifndef OPENGL_H
#define OPENGL_H

// Include GLEW. Always include it before gl.h and glfw.h
#include <GL/glew.h>

// Include GLFW to handle the window and the user input
#include <GLFW/glfw3.h>

// Include GLM, a library for 3D mathematics
#include </user/HS222/tw00275/glm-bin/include/glm/glm.hpp>
#include </user/HS222/tw00275/glm-bin/include/glm/gtc/matrix_transform.hpp>
//using namespace glm;

#include "shader.h"
#include "texture.h"
#include "objloader.h"
#include "vboindexer.h"
//#include <common/controls.hpp>

#include </user/HS222/tw00275/glm-bin/include/glm/gtc/quaternion.hpp>
#include </user/HS222/tw00275/glm-bin/include/glm/gtx/quaternion.hpp>

//#include <windows.h>	// Header File For Windows
#include <GL/gl.h>		// Header File For The OpenGL32 Library
#include <GL/glu.h>		// Header File For The GLu32 Library

#include "/user/HS222/tw00275/ceres-bin/include/ceres/ceres.h"
#include "glog/logging.h"

using ceres::AutoDiffCostFunction;
using ceres::CostFunction;
using ceres::Problem;
using ceres::Solver;
using ceres::Solve;

class OpenGL {
	public:
		OpenGL(cv::Mat cv_depth, cv::Mat cv_normals, cv::Mat cv_texture, std::vector<cv::Mat> textureImages, cv::Mat lightDirections, int windowWidth, int windowHeight, const std::string modelPath) {
			// Initialise the GLFW library first
			if (!glfwInit()) {
				std::cerr << "Failed to initialise GLFW.\n";
				getchar();
				return;
			}
			std::cout << "Initialised GLFW.\n";

			glfwWindowHint(GLFW_SAMPLES, 4); // 4x antialiasing
			glfwWindowHint(GLFW_RESIZABLE, GL_FALSE);
			glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3); // OpenGL 3.3
			glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
			glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE); // Only needed for macOS
			glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE); // Disable old OpenGL
			
			// Set the shadow resolution to 4096x4096
			int shadowResolution = 4096;

			// Open a window and create its OpenGL context
			window = glfwCreateWindow(windowWidth, windowHeight, "Window", NULL, NULL);
			if (window == NULL) {
				std::cerr << "Failed to open GLFW window. Older Intel GPUs are not OpenGL 3.3 compatible.\n";
				getchar();
				glfwTerminate();
				return;
			}
			glfwMakeContextCurrent(window);
			std::cout << "Opened GLFW window.\n";

			// On macOS with a retina screen it will be windowWidth*2 and windowHeight*2, so the actual framebuffer size is:
			glfwGetFramebufferSize(window, &windowWidth, &windowHeight);

			// Initialize GLEW
			glewExperimental = true; // Needed for core profile
			if (glewInit() != GLEW_OK) {
				std::cerr << "Failed to initialise GLEW.\n";
				getchar();
				glfwTerminate();
				return;
			}
			std::cout << "Initialised GLEW.\n";

			// Ensure that the escape key being pressed can be captured
			glfwSetInputMode(window, GLFW_STICKY_KEYS, GL_TRUE);
			// Hide the mouse and enable unlimited mouvement
			//glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_DISABLED);

			// Set the mouse at the center of the screen
			glfwPollEvents();
			//glfwSetCursorPos(window, windowWidth / 2, windowHeight / 2);

			// Dark blue background
			glClearColor(0.0f, 0.0f, 0.4f, 0.0f);

			// Enable depth test, which stores fragments in the Z-buffer
			glEnable(GL_DEPTH_TEST);

			// Accept fragment if it is closer to the camera than the former one
			glDepthFunc(GL_LESS);

			// Cull triangles which do not have a normal facing towards the camera
			glEnable(GL_CULL_FACE);

			// Create a Vertex Array Object (VAO) and set it as the current one
			//GLuint VertexArrayID;
			glGenVertexArrays(1, &VertexArrayID);
			glBindVertexArray(VertexArrayID);

			// Create and compile the GLSL program from the shaders
			depthProgramID = LoadShaders("/user/HS222/tw00275/PhD/Minimisation/apps/specular_estimation/src/DepthRTT.vertexshader", "/user/HS222/tw00275/PhD/Minimisation/apps/specular_estimation/src/DepthRTT.fragmentshader");

			// Get a handle for the "MVP" uniform
			GLuint depthMatrixID = glGetUniformLocation(depthProgramID, "depthMVP");

			// Load the texture
			//GLuint Texture = loadDDS("C:/Users/tdwal/Documents/repos/OpenGLTutorials/tutorial16_shadowmaps/uvmap.DDS");
			Texture = loadMat();
			std::cout << "Texture loaded.\n\n";

			// Read the .obj file
			//bool res = loadOBJ("C:/Users/tdwal/Documents/repos/OpenGLTutorials/tutorial16_shadowmaps/room_thickwalls.obj", vertices, uvs, normals);
			createMesh(cv_normals, cv_depth);
			std::cout << "Model created.\n\n";

			// Load it into a VBO
			//GLuint vertexbuffer; // This will identify the vertex buffer
			glGenBuffers(1, &vertexbuffer); // Generate 1 buffer, and put the resulting identifier in vertexbuffer
			glBindBuffer(GL_ARRAY_BUFFER, vertexbuffer); // The following commands will talk about our 'vertexbuffer' buffer
			glBufferData(GL_ARRAY_BUFFER, vertices.size() * sizeof(glm::vec3), &vertices[0], GL_STATIC_DRAW); // Give the vertices to OpenGL

			//GLuint uvbuffer;
			glGenBuffers(1, &uvbuffer);
			glBindBuffer(GL_ARRAY_BUFFER, uvbuffer);
			glBufferData(GL_ARRAY_BUFFER, uvs.size() * sizeof(glm::vec2), &uvs[0], GL_STATIC_DRAW);

			//GLuint normalbuffer;
			glGenBuffers(1, &normalbuffer);
			glBindBuffer(GL_ARRAY_BUFFER, normalbuffer);
			glBufferData(GL_ARRAY_BUFFER, normals.size() * sizeof(glm::vec3), &normals[0], GL_STATIC_DRAW);

			// Generate a buffer for the indices as well
			//GLuint elementbuffer;
			glGenBuffers(1, &elementbuffer);
			glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, elementbuffer);
			glBufferData(GL_ELEMENT_ARRAY_BUFFER, indices.size() * sizeof(unsigned int), &indices[0], GL_STATIC_DRAW);


			// ---------------------------------------------
			// Render to Texture - specific code begins here
			// ---------------------------------------------

			// The framebuffer, which regroups 0, 1 or more textures, and 0 or 1 depth buffers
			FramebufferName = 0;
			glGenFramebuffers(1, &FramebufferName);
			glBindFramebuffer(GL_FRAMEBUFFER, FramebufferName);

			// Depth texture. Slower than a depth buffer, but can be sampled later in the shader
			//GLuint depthTexture;
			glGenTextures(1, &depthTexture);
			glBindTexture(GL_TEXTURE_2D, depthTexture);
			glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT16, shadowResolution, shadowResolution, 0, GL_DEPTH_COMPONENT, GL_FLOAT, 0);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC, GL_LEQUAL);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_COMPARE_R_TO_TEXTURE);

			glFramebufferTexture(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, depthTexture, 0);

			// No colour output in the bound framebuffer, only depth.
			glDrawBuffer(GL_NONE);

			// Always check if there are any errors in the framebuffer
			if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
				return;

			// The quad's FBO. Used only for visualising the shadowmap.
			static const GLfloat g_quad_vertex_buffer_data[] = {
				-1.0f, -1.0f, 0.0f,
				 1.0f, -1.0f, 0.0f,
				-1.0f,  1.0f, 0.0f,
				-1.0f,  1.0f, 0.0f,
				 1.0f, -1.0f, 0.0f,
				 1.0f,  1.0f, 0.0f,
			};

			//GLuint quad_vertexbuffer;
			glGenBuffers(1, &quad_vertexbuffer);
			glBindBuffer(GL_ARRAY_BUFFER, quad_vertexbuffer);
			glBufferData(GL_ARRAY_BUFFER, sizeof(g_quad_vertex_buffer_data), g_quad_vertex_buffer_data, GL_STATIC_DRAW);

			// Create and compile the GLSL program from the shaders
			/*quad_programID = LoadShaders("/user/HS222/tw00275/PhD/Minimisation/apps/specular_estimation/src/Passthrough.vertexshader", "/user/HS222/tw00275/PhD/Minimisation/apps/specular_estimation/src/SimpleTexture.fragmentshader");
			std::cout << "Shaders loaded.\n";
			texID = glGetUniformLocation(quad_programID, "myTexture");
			std::cout << "Uniform Location of texture found.\n";*/

			// Create and compile the GLSL program from the shaders
			programID = LoadShaders("/user/HS222/tw00275/PhD/Minimisation/apps/specular_estimation/src/ShadowMapping.vertexshader", "/user/HS222/tw00275/PhD/Minimisation/apps/specular_estimation/src/ShadowMapping.fragmentshader");

			// Get a handle for the "myTextureSampler" uniform
			TextureID = glGetUniformLocation(programID, "myTextureSampler");
			//SpecularID = glGetUniformLocation(programID, "mySpecularSampler");
			
			SpecularIntensityID = glGetUniformLocation(programID, "specularIntensity");
			SpecularPowerID = glGetUniformLocation(programID, "specularPower");

			// Get a handle for the "MVP" uniform
			MatrixID = glGetUniformLocation(programID, "MVP");
			ViewMatrixID = glGetUniformLocation(programID, "V");
			ModelMatrixID = glGetUniformLocation(programID, "M");
			DepthBiasID = glGetUniformLocation(programID, "DepthBiasMVP");
			ShadowMapID = glGetUniformLocation(programID, "shadowMap");

			// Get a handle for the "LightPosition" uniform
			lightInvDirID = glGetUniformLocation(programID, "LightInvDirection_worldspace");

			// Compute the MVP matrix from the light's point of view
			lightInvDir = glm::vec3(0, 0, 1);
			depthProjectionMatrix = glm::perspective<float>(45.0f, 1.0f, 1.0f, 100.0f);
			depthViewMatrix = glm::lookAt(lightInvDir, glm::vec3(0, 0, 0), glm::vec3(0, 1, 0));

			// Initial position: on +Z
			position = glm::vec3(0, 0, 1);
			// Initial horizontal angle: toward -Z
			horizontalAngle = M_PI;
			// Initial vertical angle: none
			verticalAngle = 0.0f;
			// Initial Field of View
			FoV = 60.0f;
			// Mouse Speed
			float mouseSpeed = 0.005f;
			
			std::cout << "Initialised OpenGL\n";
		}
		
		~OpenGL() {
			// Cleanup VBO and shader
			glDeleteBuffers(1, &vertexbuffer);
			glDeleteBuffers(1, &uvbuffer);
			glDeleteBuffers(1, &normalbuffer);
			glDeleteBuffers(1, &elementbuffer);
			glDeleteProgram(programID);
			glDeleteProgram(depthProgramID);
			//glDeleteProgram(quad_programID);
			glDeleteTextures(1, &Texture);

			glDeleteFramebuffers(1, &FramebufferName);
			glDeleteTextures(1, &depthTexture);
			glDeleteBuffers(1, &quad_vertexbuffer);
			glDeleteVertexArrays(1, &VertexArrayID);

			// Close OpenGL window and terminate GLFW
			glfwTerminate();
		}
	
		void renderPolygons(double SpecularIntensity, double SpecularPower) {
			
			std::cout << "Rendering polygons...\n";
			
			// Render to the screen
			glBindFramebuffer(GL_FRAMEBUFFER, 0);
			glViewport(0, 0, windowWidth, windowHeight); // Render on the whole framebuffer, complete from the lower left corner to the upper right

			glEnable(GL_CULL_FACE);
			glCullFace(GL_BACK); // Cull back-facing triangles -> draw only front-facing triangles

			// Clear the screen
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

			// Use the shader
			glUseProgram(programID);
			
			// Send the transformation to the currently bound shader, in the "MVP" uniform
			glUniformMatrix4fv(MatrixID,      1, GL_FALSE, &MVP[0][0]);
			glUniformMatrix4fv(ModelMatrixID, 1, GL_FALSE, &ModelMatrix[0][0]);
			glUniformMatrix4fv(ViewMatrixID,  1, GL_FALSE, &ViewMatrix[0][0]);
			glUniformMatrix4fv(DepthBiasID,   1, GL_FALSE, &depthBiasMVP[0][0]);
			
			glUniform1f(SpecularIntensityID, SpecularIntensity);
			glUniform1f(SpecularPowerID,     SpecularPower);

			glUniform3f(lightInvDirID, lightInvDir.x, lightInvDir.y, lightInvDir.z);

			// Bind the texture in Texture Unit 0
			glActiveTexture(GL_TEXTURE0);
			glBindTexture(GL_TEXTURE_2D, Texture);
			// Set the "myTextureSampler" sampler to use Texture Unit 0
			glUniform1i(TextureID, 0);

			glActiveTexture(GL_TEXTURE1);
			glBindTexture(GL_TEXTURE_2D, depthTexture);
			glUniform1i(ShadowMapID, 1);

			// 1st attribute buffer: vertices
			glEnableVertexAttribArray(0);
			glBindBuffer(GL_ARRAY_BUFFER, vertexbuffer);
			glVertexAttribPointer(
				0,			// attribute
				3,			// size
				GL_FLOAT,	// type
				GL_FALSE,	// normalized?
				0,			// stride
				(void*)0	// array buffer offset
			);

			// 2nd attribute buffer: UVs
			glEnableVertexAttribArray(1);
			glBindBuffer(GL_ARRAY_BUFFER, uvbuffer);
			glVertexAttribPointer(
				1,			// attribute
				2,			// size
				GL_FLOAT,	// type
				GL_FALSE,	// normalized?
				0,			// stride
				(void*)0	// array buffer offset
			);

			// 3rd attribute buffer: normals
			glEnableVertexAttribArray(2);
			glBindBuffer(GL_ARRAY_BUFFER, normalbuffer);
			glVertexAttribPointer(
				2,			// attribute
				3,			// size
				GL_FLOAT,	// type
				GL_FALSE,	// normalized?
				0,			// stride
				(void*)0	// array buffer offset
			);

			// Index buffer
			glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, elementbuffer);

			// Draw the triangles
			glDrawElements(
				GL_TRIANGLES,				// mode
				GLsizei(indices.size()),	// count
				GL_UNSIGNED_INT,			// type
				(void*)0					// element array buffer offset
			);

			glDisableVertexAttribArray(0);
			glDisableVertexAttribArray(1);
			glDisableVertexAttribArray(2);
		}
		
		void createMesh(cv::Mat cv_normals, cv::Mat cv_depth) {
			//std::cout << "Width: " << cv_normals.cols << ", Height: " << cv_normals.rows << std::endl;
			std::cout << "Creating " << cv_normals.cols << "x" << cv_normals.rows << " mesh from depth map and normals... ";

			flip(cv_normals, cv_normals, 0);
			flip(cv_depth, cv_depth, 0);

			for (int y = 0; y <= cv_normals.rows; y++) {
				for (int x = 0; x <= cv_normals.cols; x++) {
					vertices.push_back(glm::vec3(float(x - (cv_normals.cols / 2)) / cv_normals.rows, float(y - (cv_normals.rows / 2)) / cv_normals.rows, 0));
					//vertices.push_back(glm::vec3(float(x - (cv_normals.cols / 2)) / cv_normals.rows, float(y - (cv_normals.rows / 2)) / cv_normals.rows, -cv_depth.at<float>(y, x) / (cv_normals.rows * 10)));
					uvs.push_back(glm::vec2(float(x) / cv_normals.cols, 1.0f - (float(y) / cv_normals.rows)));
					
					if (y == cv_normals.rows) {
						normals.push_back(glm::vec3(cv_normals.at<cv::Vec3f>(y-1, x)[1], -cv_normals.at<cv::Vec3f>(y-1, x)[0], cv_normals.at<cv::Vec3f>(y-1, x)[2]));
					} else if (x == cv_normals.cols) {
						normals.push_back(glm::vec3(cv_normals.at<cv::Vec3f>(y, x-1)[1], -cv_normals.at<cv::Vec3f>(y, x-1)[0], cv_normals.at<cv::Vec3f>(y, x-1)[2]));
					} else if (y == cv_normals.rows && x == cv_normals.cols) {
						normals.push_back(glm::vec3(cv_normals.at<cv::Vec3f>(y-1, x-1)[1], -cv_normals.at<cv::Vec3f>(y-1, x-1)[0], cv_normals.at<cv::Vec3f>(y-1, x-1)[2]));
					} else {
						normals.push_back(glm::vec3(cv_normals.at<cv::Vec3f>(y, x)[1], -cv_normals.at<cv::Vec3f>(y, x)[0], cv_normals.at<cv::Vec3f>(y, x)[2]));
						//normals.push_back(glm::vec3(0, 0, 1));
					}
				}
			}

			std::cout << "Mesh created with " << cv_normals.cols * cv_normals.rows * 2 << " polygons.\n";

			std::cout << "Indexing vertices... ";

			for (int y = 0; y < cv_normals.rows; y++) {
				for (int x = 0; x < cv_normals.cols; x++) {
					indices.push_back(int((y * cv_normals.cols) + x));
					indices.push_back(int((y * cv_normals.cols) + x + 1));
					indices.push_back(int(((y + 1) * cv_normals.cols) + x + 1));

					indices.push_back(int(((y + 1) * cv_normals.cols) + x + 1));
					indices.push_back(int(((y + 1) * cv_normals.cols) + x));
					indices.push_back(int((y * cv_normals.cols) + x));
				}
			}

			std::cout << "Vertices indexed.\n\n";
		}
		
		void computeMatricesFromLights() {
			//calculateResidual = false;
			
			// Direction: Convert Spherical coordinates to Cartesian coordinates
			glm::vec3 direction(
				cos(verticalAngle) * sin(horizontalAngle),
				sin(verticalAngle),
				cos(verticalAngle) * cos(horizontalAngle)
			);

			// Right vector
			glm::vec3 right = glm::vec3(
				sin(horizontalAngle - M_PI / 2.0f),
				0,
				cos(horizontalAngle - M_PI / 2.0f)
			);

			// Up vector
			glm::vec3 up = glm::cross(right, direction);
			
			lightInvDir = glm::vec3(lightDirections.at<float>(lightNumber, 1), -lightDirections.at<float>(lightNumber, 0), lightDirections.at<float>(lightNumber, 2));
			
			float aspectRatio = float(windowWidth) / float(windowHeight);

			if (perspectiveProjection) // Projection matrix: 60° Field of View, 16:9 aspect ratio, display range: 0.1 units <-> 100 units
				ProjectionMatrix = glm::perspective(glm::radians(FoV), aspectRatio, 0.1f, 100.0f);
			else
				ProjectionMatrix = glm::ortho(-0.5f * aspectRatio, 0.5f * aspectRatio, -0.5f, 0.5f, 0.0f, 10.0f); // In world coordinates

			// Camera matrix
			ViewMatrix = glm::lookAt(
				position,				// Camera is here
				position + direction,	// and looks here: at the same position, plus "direction"
				up						// Head is up (set to 0, -1, 0 to look upside-down)
			);
		}
	
		GLuint loadMat() {
			// Actual RGB data
			unsigned char * data = cv_texture.data;
			unsigned int width  = cv_texture.cols;
			unsigned int height = cv_texture.rows;

			// Create one OpenGL texture
			GLuint textureID;
			glGenTextures(1, &textureID);

			// "Bind" the newly created texture: all future texture functions will modify this texture
			glBindTexture(GL_TEXTURE_2D, textureID);

			// Give the image to OpenGL
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, GL_BGR, GL_UNSIGNED_BYTE, data);

			// OpenGL has now copied the data. Free this version
			//delete[] data;

			// Nearest neighbour filtering
			//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
			//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST); 

			// Trilinear filtering
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
			// This requires mipmaps. Generate them automatically.
			glGenerateMipmap(GL_TEXTURE_2D);

			// Return the ID of the texture we just created
			return textureID;
		}
	
		void computeResidual(double SpecularIntensity, double SpecularPower) {
			// Create an empty Mat the same size as the image
			cv::Mat temp = cv::Mat(windowHeight, windowWidth, CV_8UC3);
			cv::Mat sumOfSquaredDifferences = cv::Mat(windowHeight, windowWidth, CV_16UC3);
			
			// Read the image into the Mat as an 8-bit colour image
			glReadPixels(0, 0, windowWidth, windowHeight, GL_BGR, GL_UNSIGNED_BYTE, temp.data);
			
			// 3D models have the origin in the bottom-left corner, while images have the origin in the top-left corner. The image is flipped to convert between the two.
			cv::flip(temp, temp, 0);
			
			//temp = temp(cv::Rect(1, 1, windowWidth-2, windowHeight-2));
			//copyMakeBorder(temp, temp, 1, 1, 1, 1, cv::BORDER_REPLICATE);
			
			// Calculate the absolute differences between the model and the photograph
			cv::absdiff(temp, textureImages[imageNumber], residual);
			
			sumOfSquaredDifferences = residual;
			
			// Square each element
			multiply(sumOfSquaredDifferences, sumOfSquaredDifferences, sumOfSquaredDifferences);
			
			std::ostringstream stm;
			stm << imageNumber;
			std::string number = stm.str();
			
			// Display the residual
			//cv::namedWindow("Model", CV_WINDOW_NORMAL);
			//cv::imshow("Model", temp);
			//cv::namedWindow("Sum of Squared Differences", CV_WINDOW_NORMAL);
			//cv::imshow("Sum of Squared Differences", sumOfSquaredDifferences);
			cv::imwrite(modelPath + "model." + number + ".png", temp);
			cv::imwrite(modelPath + "residual." + number + ".png", residual);
			
			// Crop a 1-pixel border around the residual
			sumOfSquaredDifferences = sumOfSquaredDifferences(cv::Rect(1, 1, windowWidth-2, windowHeight-2));
			
			//sum = (cv::sum(sumOfSquaredDifferences)[0] + cv::sum(sumOfSquaredDifferences)[1] + cv::sum(sumOfSquaredDifferences)[2])/(windowHeight-2 * windowWidth-2);
			sum = (cv::sum(sumOfSquaredDifferences)[0] + cv::sum(sumOfSquaredDifferences)[1] + cv::sum(sumOfSquaredDifferences)[2]);
			//residuals = cv::Vec3d(cv::sum(sumOfSquaredDifferences)[0], cv::sum(sumOfSquaredDifferences)[1], cv::sum(sumOfSquaredDifferences)[2]);
			
			
			/*
			// Use the matchTemplate function to calculate the sum of squared differences
			matchTemplate(temp, textureImages[imageNumber], residual, 1);
			
			sum = cv::sum(residual)[0];
			*/
			
			std::cout << "Image " << imageNumber << ", SSD per pixel: R = " << cv::sum(sumOfSquaredDifferences)[2]/(windowHeight * windowWidth) << ", G = " << cv::sum(sumOfSquaredDifferences)[1]/(windowHeight * windowWidth) << ", B = " << cv::sum(sumOfSquaredDifferences)[0]/(windowHeight * windowWidth) << ", specular intensity = " << SpecularIntensity << ", specular power = " << SpecularPower << std::endl;
			//std::cout << "Image " << imageNumber << ", average sum of squared differences: R = " << sum << ", specular intensity = " << SpecularIntensity << ", specular power = " << SpecularPower << std::endl;
		}
		
		void computeResiduals(cv::Vec3d& residual, std::vector< cv::Vec3d >& residuals, double& totalResidual, cv::Mat& residualImage, bool calculateResidual, std::vector< cv::Mat > textureImages, int& lightNumber, int numberOfLights, double& SpecularIntensity, double& SpecularPower, std::vector<unsigned int> indices, std::string modelPath) {
			do {
				// Render the shadows
				depthModelMatrix = glm::mat4(1.0);
				depthMVP = depthProjectionMatrix * depthViewMatrix * depthModelMatrix;
				//renderShadows(FramebufferName, shadowResolution, depthProgramID, depthProjectionMatrix, depthViewMatrix, depthMatrixID, vertexbuffer, elementbuffer, indices, depthModelMatrix, depthMVP);

				// Compute the MVP matrix from keyboard and mouse input
				ModelMatrix = glm::mat4(1.0);
				
				//computeMatricesFromInputs(windowWidth, windowHeight, position, horizontalAngle, verticalAngle, FoV, mouseSpeed, ProjectionMatrix, ViewMatrix, lightInvDir, depthProjectionMatrix, depthViewMatrix, perspectiveProjection, shadowControl, SpecularIntensityID, SpecularIntensity, SpecularPowerID, SpecularPower, calculateResidual);
				computeMatricesFromLights();
				MVP = ProjectionMatrix * ViewMatrix * ModelMatrix;

				biasMatrix = glm::mat4(
					0.5, 0.0, 0.0, 0.0,
					0.0, 0.5, 0.0, 0.0,
					0.0, 0.0, 0.5, 0.0,
					0.5, 0.5, 0.5, 1.0
				);

				depthBiasMVP = biasMatrix * depthMVP;
				
					
				// Render the polygons
				renderPolygons(SpecularIntensity, SpecularPower);
				
				// Optionally render the shadowmap (for debug only)
				//void renderShadowMap(GLuint quad_programID, GLuint depthTexture, GLuint texID, GLuint quad_vertexbuffer);
				
				
				
				//totalResidual = returnResidual(lightNumber, height, width, textureImages, residualImage, residuals, SpecularIntensity, SpecularPower);
				
				//computeResidual(SpecularIntensity, SpecularPower);
				//residuals.push_back(residual);
				
				// Set up the only cost function (also known as residual). This uses auto-differentiation to obtain the derivative (jacobian).
				// AutoDiffCostFunction<CostFunctor, (Dimensions of Residual), (Dimensions of Variables)>
				//CostFunction* cost_function = new AutoDiffCostFunction<CostFunctor, 1, 2>(new CostFunctor);
				//problem.AddResidualBlock(cost_function, NULL, &SpecularIntensity, &SpecularPower);
				
				// Swap buffers
				glfwSwapBuffers(window);
				glfwPollEvents();
				
				
				
				//TODO: Save model image and residual
				
				
				/*CostFunction* cost_function = new AutoDiffCostFunction<CostFunctor, 1, 1>(new CostFunctor);
				problem.AddResidualBlock(cost_function, NULL, &x);
				*/
				lightNumber++;
			//} while (lightNumber < numberOfLights);
			} while (glfwGetKey(window, GLFW_KEY_ESCAPE) != GLFW_PRESS && glfwWindowShouldClose(window) == 0); // Check if the ESC key was pressed or the window was closed
			
			lightNumber = 0;
		}
	
	private:
		cv::Mat cv_depth;
		cv::Mat cv_normals;
		cv::Mat cv_texture;
		std::vector<cv::Mat> textureImages;
		cv::Mat lightDirections;
		int windowWidth;
		int windowHeight;
		int lightNumber;
		int imageNumber;
		std::vector<unsigned int> indices;
		cv::Vec3d residual;
		std::vector<cv::Vec3d> residuals;
		const std::string modelPath;
		
		glm::mat4 depthProjectionMatrix;
		glm::mat4 depthViewMatrix;
		glm::vec3 position;
		float horizontalAngle;
		float verticalAngle;
		float FoV;
		glm::vec3 lightInvDir;
		GLuint programID;
		GLuint MatrixID;
		GLuint ModelMatrixID;
		GLuint ViewMatrixID;
		GLuint DepthBiasID;
		GLuint lightInvDirID;
		GLuint Texture;
		GLuint TextureID;
		GLuint depthTexture;
		GLuint ShadowMapID;
		GLuint vertexbuffer;
		GLuint uvbuffer;
		GLuint normalbuffer;
		GLuint elementbuffer;
		GLuint depthProgramID;
		//GLuint quad_programID;
		GLuint FramebufferName;
		GLuint quad_vertexbuffer;
		GLuint VertexArrayID;
		GLuint texID;
		
		glm::mat4 depthModelMatrix;
		glm::mat4 depthMVP;
		glm::mat4 ModelMatrix;
		glm::mat4 MVP;
		glm::mat4 ViewMatrix;
		glm::mat4 depthBiasMVP;
		glm::mat4 ProjectionMatrix;
		glm::mat4 biasMatrix;
		bool perspectiveProjection;
		bool shadowControl;
		//glm::mat4 biasMatrix;
		
		std::vector<glm::vec3> vertices;
		std::vector<glm::vec2> uvs;
		std::vector<glm::vec3> normals;
		//std::vector<unsigned int> indices;
		
		double sum;
		
		GLuint SpecularIntensityID;
		double SpecularIntensity;
		GLuint SpecularPowerID;
		double SpecularPower;
};



// From https://codeyarns.com/2015/06/16/how-to-rotate-point-in-opencv/ 
/*cv::Mat rotatePoints(cv::Mat point, cv::Vec3d rvec) {
	cv::Mat rotatedPoints, pointsTransposed;

	cv::Mat rx = (cv::Mat_<float>(3, 3) << 1, 0, 0, 0, cos(rvec[0]), -sin(rvec[0]), 0, sin(rvec[0]), cos(rvec[0]));
	cv::Mat ry = (cv::Mat_<float>(3, 3) << cos(rvec[0]), 0, sin(rvec[0]), 0, 1, 0, -sin(rvec[0]), 0, cos(rvec[0]));
	cv::Mat rz = (cv::Mat_<float>(3, 3) << cos(rvec[0]), -sin(rvec[0]), 0, sin(rvec[0]), cos(rvec[0]), 0, 0, 0, 1);

	cv::transpose(point, pointsTransposed);

	rotatedPoints = rx * ry * rz * pointsTransposed;

	cv::transpose(rotatedPoints, point);

	return point;
}*/

// From https://codeyarns.com/2015/06/16/how-to-rotate-point-in-opencv/
/*cv::Point3f rotatePoint(const cv::Point3f& cen_pt, const cv::Point3f& p, float rad)
{
	const cv::Point3f trans_pt = p - cen_pt;
	const cv::Point3f rot_pt = RotatePoint(trans_pt, rad);
	const cv::Point3f fin_pt = rot_pt + cen_pt;

	return fin_pt;
}*/

#endif
