#include "stdafx.h"

int main(int argc, char** argv) {
	
	// Required for the Ceres solver
	google::InitGoogleLogging(argv[0]);
	
	// Create strings to store the model folder name and calibration folder name
	// By default, the name of the chrome sphere calibration images folder is "chrome". This will be overwritten if a second argument is provided.
	std::string imageName, calibration = "chrome";
	// Using an image scale of 1 retains the full resolution
	int imageScale = 1;
	// Show textue is set to false if cropping is used
	bool showTexture = true;
	
	// Read the name of the model folder
	if (argc >= 2) {
		// Read the file name from the command line argument and convert it from a char array to a string
		const char *argument1 = argv[1];
		imageName = argument1;
	}
	// Read the name of the calibration folder
	if (argc >= 3) { // The specified calibration path will be used, otherwise "chrome" is used	
		const char *argument2 = argv[2];
		calibration = argument2;
	}
	// Read the image scale
	if (argc >= 4) { // An image scale is provided	
		// Scale down the images for faster processing and rendering
		const char *argument3 = argv[3];
		std::stringstream stringValue;
		stringValue << argument3;
		stringValue >> imageScale;
		if (imageScale > 1)
			std::cout << "The images will be scaled to 1/" << imageScale << " of their original resolution in each axis." << std::endl;
		else
			imageScale = 1;
	}
	if ((argc < 2) || (argc > 4)) {
		std::cerr << "Enter the name of the object followed by the calibration set to use.\n\nThe images can be scaled down by providing a number.\nFor instance, entering 2 will scale the x and y axes by 1/2." << std::endl;
		return -1;
	}

	std::cout << std::endl;

	// Define the paths for the model images, calibration images and the albedo texture
	const std::string folder		  = "2017-12-04";
	const std::string modelPath       = "/media/tw00275/ESD-USB/" + folder + "/" + imageName   + "/" + imageName   + ".";
	const std::string calibrationPath = "/media/tw00275/ESD-USB/" + folder + "/" + calibration + "/" + calibration + ".";
	const std::string texturePath     = modelPath + "texture.png";

	// A vector of Mats allow multiple images to be allocated to the same object
	// textureImages are the full colour photographs that are used to create the texture
	// modelImages are the greyscale version of textureImages which are used to calculate the surface normals
	// calibrationImages contain the reflections of the chrome sphere, and are used to determine the light directions
	// charucoImages contain the ChArUco board, and are used to calibrate the camera
	std::vector< cv::Mat > textureImages, modelImages, calibrationImages, charucoImages;
	cv::Mat heightMap, normalMap, texture, lightDirections;
	cv::Rect calibrationBoundingBox;
	int width, height;

	// Load the textureImages, modelImages, calibrationImages and charucoImages, then determine the bounding box around the chrome sphere
	loadImages(imageName, calibration, width, height, imageScale, textureImages, calibrationPath, calibrationImages, modelPath, modelImages, calibrationBoundingBox);

	// The number of lights is equal to the number of calibration images
	int numberOfLights = int(calibrationImages.size());
	// The rotation matrix is created from the rotation and translation vectors, and describes the rotation of the ChArUco board
	// The inverse of this transformation is applied to the light directions
	// The perspective transform warps the model images to align them with the ChArUco board
	std::vector< cv::Vec3d > rvecs, tvecs;
	cv::Mat rotationMatrix, rotationVector, charucoBoardImage, cameraMatrix, distortionCoefficients, perspectiveTransform, lightDirectionsPerspective = cv::Mat(numberOfLights, 3, CV_32F);
	cv::Ptr<cv::aruco::CharucoBoard> charucoBoard;
	
	charucoCalibration(width, height, charucoBoard, charucoBoardImage, cameraMatrix, distortionCoefficients);

	//charucoAlignment(charucoImages, textureImages, numberOfLights, width, height, rotationMatrix, rotationVector, perspectiveTransform, lightDirections, lightDirectionsPerspective, rvecs, tvecs, charucoBoard, charucoBoardImage, cameraMatrix, distortionCoefficients);

/*
	rotationMatrix = 
	rotationVector = 
	perspectiveTransform = 
	lightDirectionsPerspective = 
	std::vector< cv::Vec3d > & rvecs
	std::vector< cv::Vec3d > & tvecs
*/

	photometricStereo(imageName, calibration, modelPath, imageScale, heightMap, normalMap, texture, lightDirections, lightDirectionsPerspective, textureImages, modelImages, calibrationImages, calibrationBoundingBox, width, height, rvecs, tvecs);
	//photometricStereo(imageName, calibration, modelPath, imageScale, heightMap, normalMap, texture, lightDirections, lightDirectionsPerspective, charucoImages, modelImages, calibrationImages, calibrationBoundingBox, width, height, rvecs, tvecs);

	

	
	std::vector<unsigned int> indices;
	cv::Vec3d residual;
	std::vector< cv::Vec3d > residuals;
	int lightNumber = 0;
	bool calculateResidual = false;
	double totalResidual, SpecularIntensity = 0.5d, SpecularPower = 2.0d, residualValue;
	cv::Mat residualImage;

	OpenGL Renderer = OpenGL(heightMap, normalMap, texture, textureImages, lightDirections, width, height, modelPath);
	
	//Renderer.renderPolygons(SpecularIntensity, SpecularPower);
	Renderer.computeResiduals(residual, residuals, totalResidual, residualImage, calculateResidual, textureImages, lightNumber, numberOfLights, SpecularIntensity, SpecularPower, indices, modelPath);

	//initialiseOpenGL(heightMap, normalMap, texture, textureImages, lightDirections, width, height, depthProjectionMatrix, depthViewMatrix, position, horizontalAngle, verticalAngle, FoV, lightInvDir, programID, MatrixID, ModelMatrixID, ViewMatrixID, DepthBiasID, lightInvDirID, Texture, TextureID, depthTexture, ShadowMapID, vertexbuffer, uvbuffer, normalbuffer, elementbuffer, indices, depthProgramID, quad_programID, FramebufferName, quad_vertexbuffer, VertexArrayID, SpecularIntensityID, SpecularIntensity, SpecularPowerID, SpecularPower);
	//initialiseOpenGL(heightMap, normalMap, texture, textureImages, lightDirectionsPerspective, width, height, depthProjectionMatrix, depthViewMatrix, position, horizontalAngle, verticalAngle, FoV, lightInvDir, programID, MatrixID, ModelMatrixID, ViewMatrixID, DepthBiasID, lightInvDirID, Texture, TextureID, depthTexture, ShadowMapID, vertexbuffer, uvbuffer, normalbuffer, elementbuffer, indices, depthProgramID, quad_programID, FramebufferName, quad_vertexbuffer, VertexArrayID, SpecularIntensityID, SpecularIntensity, SpecularPowerID, SpecularPower);
	//initialiseOpenGL(heightMap, normalMap, texture, charucoImages, lightDirectionsPerspective, width, height, depthProjectionMatrix, depthViewMatrix, position, horizontalAngle, verticalAngle, FoV, lightInvDir, programID, MatrixID, ModelMatrixID, ViewMatrixID, DepthBiasID, lightInvDirID, Texture, TextureID, depthTexture, ShadowMapID, vertexbuffer, uvbuffer, normalbuffer, elementbuffer, indices, depthProgramID, quad_programID, FramebufferName, quad_vertexbuffer, VertexArrayID, SpecularIntensityID, SpecularIntensity, SpecularPowerID, SpecularPower);

	
	//specularMinimisation(SpecularIntensity, SpecularPower, residualValue, residual, residuals, totalResidual, residualImage, calculateResidual, depthProjectionMatrix, depthViewMatrix, width, height, position, horizontalAngle, verticalAngle, FoV, lightInvDir, lightDirections, textureImages, lightNumber, numberOfLights, SpecularIntensityID, SpecularPowerID, programID, ModelMatrixID, ViewMatrixID, DepthBiasID, lightInvDirID, Texture, TextureID, depthTexture, ShadowMapID, vertexbuffer, uvbuffer, normalbuffer, elementbuffer, indices, MatrixID, modelPath);

	/*
	// The variable to solve for with its initial value. It will be mutated in place by the solver.
	double x = SpecularIntensity;
	const double initial_x = x;

	// Build the problem.
	ceres::Problem problem;
	
	
	computeResiduals(residual, residuals, totalResidual, residualImage, calculateResidual, depthProjectionMatrix, depthViewMatrix, width, height, position, horizontalAngle, verticalAngle, FoV, lightInvDir, lightDirections, textureImages, lightNumber, numberOfLights, SpecularIntensityID, SpecularIntensity, SpecularPowerID, SpecularPower, programID, ModelMatrixID, ViewMatrixID, DepthBiasID, lightInvDirID, Texture, TextureID, depthTexture, ShadowMapID, vertexbuffer, uvbuffer, normalbuffer, elementbuffer, indices, MatrixID, modelPath, x, problem);
	
	
	// Run the solver!
	ceres::Solver::Options options;
	options.minimizer_progress_to_stdout = true;
	ceres::Solver::Summary summary;
	Solve(options, &problem, &summary);

	std::cout << summary.BriefReport() << "\n";
	std::cout << "Specular Intensity: " << initial_x << " -> " << x << "\n";
	*/
	
	//viewModel(residual, residuals, totalResidual, residualImage, calculateResidual, depthProjectionMatrix, depthViewMatrix, width, height, position, horizontalAngle, verticalAngle, FoV, lightInvDir, lightDirections, textureImages, lightNumber, numberOfLights, SpecularIntensityID, SpecularIntensity, SpecularPowerID, SpecularPower, programID, ModelMatrixID, ViewMatrixID, DepthBiasID, lightInvDirID, Texture, TextureID, depthTexture, ShadowMapID, vertexbuffer, uvbuffer, normalbuffer, elementbuffer, indices, MatrixID, modelPath);
	
	//terminateOpenGL(vertexbuffer, uvbuffer, normalbuffer, elementbuffer, programID, depthProgramID, quad_programID, Texture, FramebufferName, depthTexture, quad_vertexbuffer, VertexArrayID);
	
	return 0;
}

// wood6: Average residual = 8.92126e+08, difference = -27795, specular intensity = 0.234, specular power = 2
	
	/* Image 0, SSD per pixel: R = 241.427, G = 232.569, B = 204.216, specular intensity = 0.237, specular power = 2
	Image 1, SSD per pixel: R = 245.135, G = 233.578, B = 197.033, specular intensity = 0.236, specular power = 2
	Image 2, SSD per pixel: R = 203.583, G = 179.558, B = 102.644, specular intensity = 0.236, specular power = 2
	Image 3, SSD per pixel: R = 227.27, G = 211.239, B = 156.883, specular intensity = 0.236, specular power = 2
	Image 4, SSD per pixel: R = 246.922, G = 237.717, B = 212.2, specular intensity = 0.236, specular power = 2
	Image 5, SSD per pixel: R = 238.288, G = 227.53, B = 172.634, specular intensity = 0.236, specular power = 2
	Average residual = 8.92122e+08, specular intensity = 0.236, specular power = 2
	*/
	
	// shirt6: Average residual = 6.80489e+08, difference = -38358, specular intensity = 0.0450001, specular power = 2
	
	/*
	Image 0, SSD per pixel: R = 207.33, G = 224.109, B = 205.311, specular intensity = 0.046, specular power = 2
	Image 1, SSD per pixel: R = 141.999, G = 174.63, B = 149.679, specular intensity = 0.045, specular power = 2
	Image 2, SSD per pixel: R = 108.538, G = 136.126, B = 115.182, specular intensity = 0.045, specular power = 2
	Image 3, SSD per pixel: R = 124.982, G = 151.519, B = 122.994, specular intensity = 0.045, specular power = 2
	Image 4, SSD per pixel: R = 166.409, G = 191.791, B = 174.871, specular intensity = 0.045, specular power = 2
	Image 5, SSD per pixel: R = 150.494, G = 183.359, B = 153.144, specular intensity = 0.045, specular power = 2
	Average residual = 6.80489e+08, specular intensity = 0.045, specular power = 2
	*/
	
	// greenpaper6: Average residual = 8.46285e+08, difference = -12212, specular intensity = 0.177, specular power = 2
	
	/*
	Image 0, SSD per pixel: R = 202.431, G = 235.015, B = 217.53, specular intensity = 0.176, specular power = 2
	Image 1, SSD per pixel: R = 202.585, G = 233.121, B = 201.261, specular intensity = 0.177, specular power = 2
	Image 2, SSD per pixel: R = 138.306, G = 204.308, B = 138.846, specular intensity = 0.177, specular power = 2
	Image 3, SSD per pixel: R = 209.332, G = 224.895, B = 163.56, specular intensity = 0.177, specular power = 2
	Image 4, SSD per pixel: R = 227.199, G = 238.421, B = 218.823, specular intensity = 0.177, specular power = 2
	Image 5, SSD per pixel: R = 171.797, G = 234.565, B = 199.287, specular intensity = 0.177, specular power = 2
	Average residual = 8.46284e+08, specular intensity = 0.177, specular power = 2
	*/
	
	// plastic6: Average residual = 9.30516e+08, difference = -30793, specular intensity = 0.42, specular power = 2
	
	/*
	Image 0, SSD per pixel: R = 231.049, G = 234.629, B = 234.937, specular intensity = 0.419, specular power = 2
	Image 1, SSD per pixel: R = 221.101, G = 218.518, B = 216.063, specular intensity = 0.42, specular power = 2
	Image 2, SSD per pixel: R = 219.93, G = 217.467, B = 217.416, specular intensity = 0.42, specular power = 2
	Image 3, SSD per pixel: R = 241.148, G = 233.661, B = 225.648, specular intensity = 0.42, specular power = 2
	Image 4, SSD per pixel: R = 240.902, G = 233.476, B = 225.837, specular intensity = 0.42, specular power = 2
	Image 5, SSD per pixel: R = 219.19, G = 223.315, B = 223.424, specular intensity = 0.42, specular power = 2
	Average residual = 9.30516e+08, specular intensity = 0.42, specular power = 2
	*/
